"""
Definitions imported from PKCS11 C headers.
"""

from cython.view cimport array

from pkcs11.exceptions import *

cdef extern from '../extern/cryptoki.h':

    ctypedef unsigned char CK_BYTE
    ctypedef CK_BYTE CK_BBOOL
    ctypedef CK_BYTE CK_UTF8CHAR
    ctypedef unsigned char CK_CHAR
    ctypedef unsigned long int CK_ULONG
    ctypedef CK_ULONG CK_ATTRIBUTE_TYPE
    ctypedef CK_ULONG CK_EC_KDF_TYPE
    ctypedef CK_ULONG CK_FLAGS
    ctypedef CK_ULONG CK_MECHANISM_TYPE
    ctypedef CK_ULONG CK_OBJECT_HANDLE
    ctypedef CK_ULONG CK_RSA_PKCS_MGF_TYPE
    ctypedef CK_ULONG CK_RSA_PKCS_OAEP_SOURCE_TYPE
    ctypedef CK_ULONG CK_SESSION_HANDLE
    ctypedef CK_ULONG CK_SLOT_ID
    ctypedef CK_ULONG CK_STATE

    ctypedef enum CK_RV:
        CKR_OK,
        CKR_CANCEL,
        CKR_HOST_MEMORY,
        CKR_SLOT_ID_INVALID,

        CKR_GENERAL_ERROR,
        CKR_FUNCTION_FAILED,

        CKR_ARGUMENTS_BAD,
        CKR_NO_EVENT,
        CKR_NEED_TO_CREATE_THREADS,
        CKR_CANT_LOCK,

        CKR_ATTRIBUTE_READ_ONLY,
        CKR_ATTRIBUTE_SENSITIVE,
        CKR_ATTRIBUTE_TYPE_INVALID,
        CKR_ATTRIBUTE_VALUE_INVALID,
        CKR_DATA_INVALID,
        CKR_DATA_LEN_RANGE,
        CKR_DEVICE_ERROR,
        CKR_DEVICE_MEMORY,
        CKR_DEVICE_REMOVED,
        CKR_ENCRYPTED_DATA_INVALID,
        CKR_ENCRYPTED_DATA_LEN_RANGE,
        CKR_FUNCTION_CANCELED,
        CKR_FUNCTION_NOT_PARALLEL,

        CKR_FUNCTION_NOT_SUPPORTED,

        CKR_KEY_HANDLE_INVALID,

        CKR_KEY_SIZE_RANGE,
        CKR_KEY_TYPE_INCONSISTENT,

        CKR_KEY_NOT_NEEDED,
        CKR_KEY_CHANGED,
        CKR_KEY_NEEDED,
        CKR_KEY_INDIGESTIBLE,
        CKR_KEY_FUNCTION_NOT_PERMITTED,
        CKR_KEY_NOT_WRAPPABLE,
        CKR_KEY_UNEXTRACTABLE,

        CKR_MECHANISM_INVALID,
        CKR_MECHANISM_PARAM_INVALID,

        CKR_OBJECT_HANDLE_INVALID,
        CKR_OPERATION_ACTIVE,
        CKR_OPERATION_NOT_INITIALIZED,
        CKR_PIN_INCORRECT,
        CKR_PIN_INVALID,
        CKR_PIN_LEN_RANGE,

        CKR_PIN_EXPIRED,
        CKR_PIN_LOCKED,

        CKR_SESSION_CLOSED,
        CKR_SESSION_COUNT,
        CKR_SESSION_HANDLE_INVALID,
        CKR_SESSION_PARALLEL_NOT_SUPPORTED,
        CKR_SESSION_READ_ONLY,
        CKR_SESSION_EXISTS,

        CKR_SESSION_READ_ONLY_EXISTS,
        CKR_SESSION_READ_WRITE_SO_EXISTS,

        CKR_SIGNATURE_INVALID,
        CKR_SIGNATURE_LEN_RANGE,
        CKR_TEMPLATE_INCOMPLETE,
        CKR_TEMPLATE_INCONSISTENT,
        CKR_TOKEN_NOT_PRESENT,
        CKR_TOKEN_NOT_RECOGNIZED,
        CKR_TOKEN_WRITE_PROTECTED,
        CKR_UNWRAPPING_KEY_HANDLE_INVALID,
        CKR_UNWRAPPING_KEY_SIZE_RANGE,
        CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT,
        CKR_USER_ALREADY_LOGGED_IN,
        CKR_USER_NOT_LOGGED_IN,
        CKR_USER_PIN_NOT_INITIALIZED,
        CKR_USER_TYPE_INVALID,

        CKR_USER_ANOTHER_ALREADY_LOGGED_IN,
        CKR_USER_TOO_MANY_TYPES,

        CKR_WRAPPED_KEY_INVALID,
        CKR_WRAPPED_KEY_LEN_RANGE,
        CKR_WRAPPING_KEY_HANDLE_INVALID,
        CKR_WRAPPING_KEY_SIZE_RANGE,
        CKR_WRAPPING_KEY_TYPE_INCONSISTENT,
        CKR_RANDOM_SEED_NOT_SUPPORTED,

        CKR_RANDOM_NO_RNG,

        CKR_DOMAIN_PARAMS_INVALID,

        CKR_BUFFER_TOO_SMALL,
        CKR_SAVED_STATE_INVALID,
        CKR_INFORMATION_SENSITIVE,
        CKR_STATE_UNSAVEABLE,

        CKR_CRYPTOKI_NOT_INITIALIZED,
        CKR_CRYPTOKI_ALREADY_INITIALIZED,
        CKR_MUTEX_BAD,
        CKR_MUTEX_NOT_LOCKED,

        CKR_NEW_PIN_MODE,
        CKR_NEXT_OTP,
        CKR_EXCEEDED_MAX_ITERATIONS,
        CKR_FIPS_SELF_TEST_FAILED,
        CKR_LIBRARY_LOAD_FAILED,
        CKR_PIN_TOO_WEAK,
        CKR_PUBLIC_KEY_INVALID,

        CKR_FUNCTION_REJECTED,

        CKR_OPERATION_CANCEL_FAILED,

        CKR_VENDOR_DEFINED,


    ctypedef enum CK_USER_TYPE:
        CKU_SO,
        CKU_USER,
        CKU_CONTEXT_SPECIFIC,
        CKU_USER_NOBODY,

    cdef enum:
        CK_TRUE,
        CK_FALSE,

    cdef enum:
        CK_UNAVAILABLE_INFORMATION,
        CK_EFFECTIVELY_INFINITE,

    cdef enum:  # CK_FLAGS
        CKF_DONT_BLOCK,
        CKF_RW_SESSION,
        CKF_SERIAL_SESSION,

    cdef enum:  # CKZ
        CKZ_DATA_SPECIFIED,

    cdef enum: # CK_STATE
        CKS_RO_PUBLIC_SESSION,
        CKS_RO_USER_FUNCTIONS,
        CKS_RW_PUBLIC_SESSION,
        CKS_RW_USER_FUNCTIONS,
        CKS_RW_SO_FUNCTIONS

    ctypedef struct CK_VERSION:
        CK_BYTE major
        CK_BYTE minor

    ctypedef struct CK_INFO:
        CK_VERSION cryptokiVersion;
        CK_UTF8CHAR manufacturerID[32]
        CK_FLAGS flags

        CK_UTF8CHAR libraryDescription[32]
        CK_VERSION libraryVersion;

    ctypedef struct CK_SLOT_INFO:
        CK_UTF8CHAR slotDescription[64]
        CK_UTF8CHAR manufacturerID[32]
        CK_FLAGS flags

        CK_VERSION hardwareVersion
        CK_VERSION firmwareVersion

    ctypedef struct CK_MECHANISM_INFO:
        CK_ULONG ulMinKeySize
        CK_ULONG ulMaxKeySize
        CK_FLAGS flags

    ctypedef struct CK_TOKEN_INFO:
        CK_UTF8CHAR label[32]
        CK_UTF8CHAR manufacturerID[32]
        CK_UTF8CHAR model[16]
        CK_CHAR serialNumber[16]
        CK_FLAGS flags

        CK_ULONG ulMaxSessionCount
        CK_ULONG ulSessionCount
        CK_ULONG ulMaxRwSessionCount
        CK_ULONG ulRwSessionCount
        CK_ULONG ulMaxPinLen
        CK_ULONG ulMinPinLen
        CK_ULONG ulTotalPublicMemory
        CK_ULONG ulFreePublicMemory
        CK_ULONG ulTotalPrivateMemory
        CK_ULONG ulFreePrivateMemory
        CK_VERSION hardwareVersion
        CK_VERSION firmwareVersion
        CK_CHAR utcTime[16]

    ctypedef struct CK_SESSION_INFO:
        CK_SLOT_ID slotID
        CK_STATE state
        CK_FLAGS flags
        CK_ULONG ulDeviceError

    ctypedef struct CK_MECHANISM:
        CK_MECHANISM_TYPE mechanism
        void *pParameter
        CK_ULONG ulParameterLen

    ctypedef struct CK_ATTRIBUTE:
        CK_ATTRIBUTE_TYPE type
        void *pValue
        CK_ULONG ulValueLen

    ctypedef struct CK_RSA_PKCS_OAEP_PARAMS:
        CK_MECHANISM_TYPE hashAlg
        CK_RSA_PKCS_MGF_TYPE mgf
        CK_RSA_PKCS_OAEP_SOURCE_TYPE source
        void *pSourceData
        CK_ULONG ulSourceDataLen

    ctypedef struct CK_RSA_PKCS_PSS_PARAMS:
        CK_MECHANISM_TYPE hashAlg
        CK_RSA_PKCS_MGF_TYPE mgf
        CK_ULONG sLen

    ctypedef struct CK_ECDH1_DERIVE_PARAMS:
        CK_EC_KDF_TYPE kdf
        CK_ULONG ulSharedDataLen
        CK_BYTE *pSharedData
        CK_ULONG ulPublicDataLen
        CK_BYTE *pPublicData

    ctypedef struct CK_AES_CTR_PARAMS:
        CK_ULONG ulCounterBits
        CK_BYTE[16] cb

    ctypedef struct CK_GCM_PARAMS:
        CK_BYTE *pIv
        CK_ULONG ulIvLen
        CK_ULONG ulIvBits
        CK_BYTE *pAAD
        CK_ULONG ulAADLen
        CK_ULONG ulTagBits

    ctypedef struct CK_KEY_DERIVATION_STRING_DATA:
        CK_BYTE *pData
        CK_ULONG ulLen

    ctypedef struct CK_AES_CBC_ENCRYPT_DATA_PARAMS:
        CK_BYTE iv[16]
        CK_BYTE *pData
        CK_ULONG length

    ctypedef struct CK_EDDSA_PARAMS:
       CK_BBOOL phFlag
       CK_ULONG ulContextDataLen
       CK_BYTE *pContextData

    cdef struct CK_FUNCTION_LIST:
        CK_VERSION version
        ## pointers to library functions are stored here
        ## caution: order matters!

        ## general purpose
        CK_RV C_Initialize(void *) nogil

        CK_RV C_Finalize(void *) nogil

        CK_RV C_GetInfo(CK_INFO *info) nogil

        CK_RV C_GetFunctionList(CK_FUNCTION_LIST **) nogil

        ## slot and token management
        CK_RV C_GetSlotList(CK_BBOOL tokenPresent,
                            CK_SLOT_ID *slotList,
                            CK_ULONG *count) nogil

        CK_RV C_GetSlotInfo(CK_SLOT_ID slotID,
                            CK_SLOT_INFO *info) nogil

        CK_RV C_GetTokenInfo(CK_SLOT_ID slotID,
                             CK_TOKEN_INFO *info) nogil

        CK_RV C_GetMechanismList(CK_SLOT_ID slotID,
                                 CK_MECHANISM_TYPE *mechanismList,
                                 CK_ULONG *count) nogil

        CK_RV C_GetMechanismInfo(CK_SLOT_ID slotID,
                                 CK_MECHANISM_TYPE mechanism,
                                 CK_MECHANISM_INFO *info) nogil

        CK_RV C_InitToken(CK_SLOT_ID slotID,
                          CK_UTF8CHAR *pPin,
                          CK_ULONG ulPinLen,
                          CK_UTF8CHAR *pLabel) nogil

        CK_RV C_InitPIN(CK_SESSION_HANDLE hSession,
                        CK_UTF8CHAR *pPin,
                        CK_ULONG ulPinLen) nogil

        CK_RV C_SetPIN(CK_SESSION_HANDLE hSession,
                       CK_UTF8CHAR *pOldPin,
                       CK_ULONG ulOldLen,
                       CK_UTF8CHAR *pNewPin,
                       CK_ULONG ulNewLen) nogil

        ## session management
        CK_RV C_OpenSession(CK_SLOT_ID slotID,
                            CK_FLAGS flags,
                            void *application,
                            void *notify,
                            CK_SESSION_HANDLE *handle) nogil

        CK_RV C_CloseSession(CK_SESSION_HANDLE session) nogil

        CK_RV C_CloseAllSessions(CK_SLOT_ID slotID) nogil

        CK_RV C_GetSessionInfo(CK_SESSION_HANDLE hSession,
                               CK_SESSION_INFO *pInfo) nogil

        CK_RV C_GetOperationState(CK_SESSION_HANDLE hSession,
                                  CK_BYTE *pOperationState,
                                  CK_ULONG *pulOperationStateLen) nogil

        CK_RV C_SetOperationState(CK_SESSION_HANDLE hSession,
                                  CK_BYTE *pOperationState,
                                  CK_ULONG ulOperationStateLen,
                                  CK_OBJECT_HANDLE hEncryptionKey,
                                  CK_OBJECT_HANDLE hAuthenticationKey) nogil

        CK_RV C_Login(CK_SESSION_HANDLE session,
                      CK_USER_TYPE userType,
                      CK_UTF8CHAR *pin,
                      CK_ULONG pinLen) nogil

        CK_RV C_Logout(CK_SESSION_HANDLE session) nogil

        ## object management
        CK_RV C_CreateObject(CK_SESSION_HANDLE session,
                             CK_ATTRIBUTE *template,
                             CK_ULONG count,
                             CK_OBJECT_HANDLE *key) nogil

        CK_RV C_CopyObject(CK_SESSION_HANDLE session,
                           CK_OBJECT_HANDLE key,
                           CK_ATTRIBUTE *template,
                           CK_ULONG count,
                           CK_OBJECT_HANDLE *new_key) nogil

        CK_RV C_DestroyObject(CK_SESSION_HANDLE session,
                              CK_OBJECT_HANDLE key) nogil

        CK_RV C_GetObjectSize(CK_SESSION_HANDLE hSession,
                              CK_OBJECT_HANDLE hObject,
                              CK_ULONG *pulSize) nogil

        CK_RV C_GetAttributeValue(CK_SESSION_HANDLE session,
                                  CK_OBJECT_HANDLE key,
                                  CK_ATTRIBUTE *template,
                                  CK_ULONG count) nogil

        CK_RV C_SetAttributeValue(CK_SESSION_HANDLE session,
                                  CK_OBJECT_HANDLE key,
                                  CK_ATTRIBUTE *template,
                                  CK_ULONG count) nogil

        CK_RV C_FindObjectsInit(CK_SESSION_HANDLE session,
                                CK_ATTRIBUTE *template,
                                CK_ULONG count) nogil

        CK_RV C_FindObjects(CK_SESSION_HANDLE session,
                            CK_OBJECT_HANDLE *objects,
                            CK_ULONG objectsMax,
                            CK_ULONG *objectsLength) nogil

        CK_RV C_FindObjectsFinal(CK_SESSION_HANDLE session) nogil

        ## encryption and decryption
        CK_RV C_EncryptInit(CK_SESSION_HANDLE session,
                            CK_MECHANISM *mechanism,
                            CK_OBJECT_HANDLE key) nogil

        CK_RV C_Encrypt(CK_SESSION_HANDLE session,
                        CK_BYTE *plaintext,
                        CK_ULONG plaintext_len,
                        CK_BYTE *ciphertext,
                        CK_ULONG *ciphertext_len) nogil

        CK_RV C_EncryptUpdate(CK_SESSION_HANDLE session,
                              CK_BYTE *part_in,
                              CK_ULONG part_in_len,
                              CK_BYTE *part_out,
                              CK_ULONG *part_out_len) nogil

        CK_RV C_EncryptFinal(CK_SESSION_HANDLE session,
                             CK_BYTE *part_out,
                             CK_ULONG *part_out_len) nogil

        CK_RV C_DecryptInit(CK_SESSION_HANDLE session,
                            CK_MECHANISM *mechanism,
                            CK_OBJECT_HANDLE key) nogil

        CK_RV C_Decrypt(CK_SESSION_HANDLE session,
                        CK_BYTE *ciphertext,
                        CK_ULONG ciphertext_len,
                        CK_BYTE *plaintext,
                        CK_ULONG *plaintext_len) nogil

        CK_RV C_DecryptUpdate(CK_SESSION_HANDLE session,
                              CK_BYTE *part_in,
                              CK_ULONG part_in_len,
                              CK_BYTE *part_out,
                              CK_ULONG *part_out_len) nogil

        CK_RV C_DecryptFinal(CK_SESSION_HANDLE session,
                             CK_BYTE *part_out,
                             CK_ULONG *part_out_len) nogil

        ## Message digests
        CK_RV C_DigestInit(CK_SESSION_HANDLE session,
                           CK_MECHANISM *mechanism) nogil

        CK_RV C_Digest(CK_SESSION_HANDLE session,
                       CK_BYTE *data,
                       CK_ULONG data_len,
                       CK_BYTE *digest,
                       CK_ULONG *digest_len) nogil

        CK_RV C_DigestUpdate(CK_SESSION_HANDLE session,
                             CK_BYTE *data,
                             CK_ULONG data_len) nogil

        CK_RV C_DigestKey(CK_SESSION_HANDLE session,
                          CK_OBJECT_HANDLE key) nogil

        CK_RV C_DigestFinal(CK_SESSION_HANDLE session,
                            CK_BYTE *digest,
                            CK_ULONG *digest_len) nogil

        ## Signing and MACing
        CK_RV C_SignInit(CK_SESSION_HANDLE session,
                         CK_MECHANISM *mechanism,
                         CK_OBJECT_HANDLE key) nogil

        CK_RV C_Sign(CK_SESSION_HANDLE session,
                     CK_BYTE *text,
                     CK_ULONG text_len,
                     CK_BYTE *signature,
                     CK_ULONG *sig_len) nogil

        CK_RV C_SignUpdate(CK_SESSION_HANDLE session,
                           CK_BYTE *part,
                           CK_ULONG part_len) nogil

        CK_RV C_SignFinal(CK_SESSION_HANDLE session,
                          CK_BYTE *signature,
                          CK_ULONG *sig_len) nogil

        CK_RV C_SignRecoverInit(CK_SESSION_HANDLE session,
                                CK_MECHANISM *mechanism,
                                CK_OBJECT_HANDLE key) nogil

        CK_RV C_SignRecover(CK_SESSION_HANDLE session,
                            CK_BYTE *text,
                            CK_ULONG text_len,
                            CK_BYTE *signature,
                            CK_ULONG *sig_len) nogil


        ## Verifying signatures and MACs
        CK_RV C_VerifyInit(CK_SESSION_HANDLE session,
                           CK_MECHANISM *mechanism,
                           CK_OBJECT_HANDLE key) nogil

        CK_RV C_Verify(CK_SESSION_HANDLE session,
                       CK_BYTE *text,
                       CK_ULONG text_len,
                       CK_BYTE *signature,
                       CK_ULONG sig_len) nogil

        CK_RV C_VerifyUpdate(CK_SESSION_HANDLE session,
                             CK_BYTE *text,
                             CK_ULONG text_len) nogil

        CK_RV C_VerifyFinal(CK_SESSION_HANDLE session,
                            CK_BYTE *signature,
                            CK_ULONG sig_len) nogil

        CK_RV C_VerifyRecoverInit(CK_SESSION_HANDLE session,
                                  CK_MECHANISM *mechanism,
                                  CK_OBJECT_HANDLE key) nogil

        CK_RV C_VerifyRecover(CK_SESSION_HANDLE session,
                              CK_BYTE *text,
                              CK_ULONG text_len,
                              CK_BYTE *signature,
                              CK_ULONG sig_len) nogil

        ## dual-function crypto operations
        CK_RV C_DigestEncryptUpdate(CK_SESSION_HANDLE session,
                                    CK_BYTE *data,
                                    CK_ULONG data_len,
                                    CK_BYTE *encrypted,
                                    CK_ULONG *encrypted_len) nogil

        CK_RV C_DecryptDigestUpdate(CK_SESSION_HANDLE session,
                                    CK_BYTE *encrypted,
                                    CK_ULONG encrypted_len,
                                    CK_BYTE *data,
                                    CK_ULONG *data_len) nogil

        CK_RV C_SignEncryptUpdate(CK_SESSION_HANDLE session,
                                  CK_BYTE *part,
                                  CK_ULONG part_len) nogil

        CK_RV C_DecryptVerifyUpdate(CK_SESSION_HANDLE session,
                                    CK_BYTE *text,
                                    CK_ULONG text_len) nogil

        ## key management
        CK_RV C_GenerateKey(CK_SESSION_HANDLE session,
                            CK_MECHANISM *mechanism,
                            CK_ATTRIBUTE *template,
                            CK_ULONG count,
                            CK_OBJECT_HANDLE *key) nogil

        CK_RV C_GenerateKeyPair(CK_SESSION_HANDLE session,
                                CK_MECHANISM *mechanism,
                                CK_ATTRIBUTE *public_template,
                                CK_ULONG public_count,
                                CK_ATTRIBUTE *private_template,
                                CK_ULONG private_count,
                                CK_OBJECT_HANDLE *public_key,
                                CK_OBJECT_HANDLE *private_key) nogil

        CK_RV C_WrapKey(CK_SESSION_HANDLE session,
                        CK_MECHANISM *mechanism,
                        CK_OBJECT_HANDLE wrapping_key,
                        CK_OBJECT_HANDLE key_to_wrap,
                        CK_BYTE *wrapped_key,
                        CK_ULONG *wrapped_key_len) nogil

        CK_RV C_UnwrapKey(CK_SESSION_HANDLE session,
                          CK_MECHANISM *mechanism,
                          CK_OBJECT_HANDLE unwrapping_key,
                          CK_BYTE *wrapped_key,
                          CK_ULONG wrapped_key_len,
                          CK_ATTRIBUTE *attrs,
                          CK_ULONG attr_len,
                          CK_OBJECT_HANDLE *unwrapped_key) nogil

        CK_RV C_DeriveKey(CK_SESSION_HANDLE session,
                          CK_MECHANISM *mechanism,
                          CK_OBJECT_HANDLE src_key,
                          CK_ATTRIBUTE *template,
                          CK_ULONG count,
                          CK_OBJECT_HANDLE *new_key) nogil

        ## random number generation
        CK_RV C_SeedRandom(CK_SESSION_HANDLE session,
                           CK_BYTE *seed,
                           CK_ULONG length) nogil

        CK_RV C_GenerateRandom(CK_SESSION_HANDLE session,
                               CK_BYTE *random,
                               CK_ULONG length) nogil


        ## parallel processing
        CK_RV C_GetFunctionStatus(CK_SESSION_HANDLE session) nogil

        CK_RV C_CancelFunction(CK_SESSION_HANDLE session) nogil

        ## smart card events
        CK_RV C_WaitForSlotEvent(CK_FLAGS flags,
                                 CK_SLOT_ID *slot,
                                 void *pRserved) nogil

# The only external API call that must be defined in a PKCS#11 library
# All other APIs are taken from the CK_FUNCTION_LIST table
ctypedef CK_RV (*C_GetFunctionList_ptr) (CK_FUNCTION_LIST **) nogil

ctypedef CK_RV (*KeyOperationInit) (
        CK_SESSION_HANDLE session,
        CK_MECHANISM *mechanism,
        CK_OBJECT_HANDLE key
) nogil
ctypedef CK_RV (*OperationUpdateWithResult) (
        CK_SESSION_HANDLE session,
        CK_BYTE *part_in,
        CK_ULONG part_in_len,
        CK_BYTE *part_out,
        CK_ULONG *part_out_len
) nogil
ctypedef CK_RV (*OperationUpdate) (
        CK_SESSION_HANDLE session,
        CK_BYTE *part_in,
        CK_ULONG part_in_len
) nogil
ctypedef CK_RV (*OperationWithResult) (
        CK_SESSION_HANDLE session,
        CK_BYTE *part_out,
        CK_ULONG *part_out_len
) nogil

cdef inline CK_BYTE_buffer(length):
    """Make a buffer for `length` CK_BYTEs."""
    return array(shape=(length,), itemsize=sizeof(CK_BYTE), format='B')


cdef inline CK_ULONG_buffer(length):
    """Make a buffer for `length` CK_ULONGs."""
    return array(shape=(length,), itemsize=sizeof(CK_ULONG), format='L')


# Note: this `cdef inline` declaration doesn't seem to be consistently labelled
# as executed by Cython's line tracing, so we flag it as nocover
# to avoid noise in the metrics.

cdef inline object map_rv_to_error(CK_RV rv):  # pragma: nocover
    if rv == CKR_ATTRIBUTE_TYPE_INVALID:
        exc = AttributeTypeInvalid()
    elif rv == CKR_ATTRIBUTE_VALUE_INVALID:
        exc = AttributeValueInvalid()
    elif rv == CKR_ATTRIBUTE_READ_ONLY:
        exc = AttributeReadOnly()
    elif rv == CKR_ATTRIBUTE_SENSITIVE:
        exc = AttributeSensitive()
    elif rv == CKR_ARGUMENTS_BAD:
        exc = ArgumentsBad()
    elif rv == CKR_BUFFER_TOO_SMALL:
        exc = PKCS11Error("Buffer was too small. Should never see this.")
    elif rv == CKR_CRYPTOKI_ALREADY_INITIALIZED:
        exc = PKCS11Error("Initialisation error (already initialized). Should never see this.")
    elif rv == CKR_CRYPTOKI_NOT_INITIALIZED:
        exc = PKCS11Error("Initialisation error (not initialized). Should never see this.")
    elif rv == CKR_DATA_INVALID:
        exc = DataInvalid()
    elif rv == CKR_DATA_LEN_RANGE:
        exc = DataLenRange()
    elif rv == CKR_DOMAIN_PARAMS_INVALID:
        exc = DomainParamsInvalid()
    elif rv == CKR_DEVICE_ERROR:
        exc = DeviceError()
    elif rv == CKR_DEVICE_MEMORY:
        exc = DeviceMemory()
    elif rv == CKR_DEVICE_REMOVED:
        exc = DeviceRemoved()
    elif rv == CKR_ENCRYPTED_DATA_INVALID:
        exc = EncryptedDataInvalid()
    elif rv == CKR_ENCRYPTED_DATA_LEN_RANGE:
        exc = EncryptedDataLenRange()
    elif rv == CKR_EXCEEDED_MAX_ITERATIONS:
        exc = ExceededMaxIterations()
    elif rv == CKR_FUNCTION_CANCELED:
        exc = FunctionCancelled()
    elif rv == CKR_FUNCTION_FAILED:
        exc = FunctionFailed()
    elif rv == CKR_FUNCTION_REJECTED:
        exc = FunctionRejected()
    elif rv == CKR_FUNCTION_NOT_SUPPORTED:
        exc = FunctionNotSupported()
    elif rv == CKR_KEY_HANDLE_INVALID:
        exc = KeyHandleInvalid()
    elif rv == CKR_KEY_INDIGESTIBLE:
        exc = KeyIndigestible()
    elif rv == CKR_KEY_NEEDED:
        exc = KeyNeeded()
    elif rv == CKR_KEY_NOT_NEEDED:
        exc = KeyNotNeeded()
    elif rv == CKR_KEY_SIZE_RANGE:
        exc = KeySizeRange()
    elif rv == CKR_KEY_NOT_WRAPPABLE:
        exc = KeyNotWrappable()
    elif rv == CKR_KEY_TYPE_INCONSISTENT:
        exc = KeyTypeInconsistent()
    elif rv == CKR_KEY_UNEXTRACTABLE:
        exc = KeyUnextractable()
    elif rv == CKR_GENERAL_ERROR:
        exc = GeneralError()
    elif rv == CKR_HOST_MEMORY:
        exc = HostMemory()
    elif rv == CKR_MECHANISM_INVALID:
        exc = MechanismInvalid()
    elif rv == CKR_MECHANISM_PARAM_INVALID:
        exc = MechanismParamInvalid()
    elif rv == CKR_NO_EVENT:
        exc = NoEvent()
    elif rv == CKR_OBJECT_HANDLE_INVALID:
        exc = ObjectHandleInvalid()
    elif rv == CKR_OPERATION_ACTIVE:
        exc = OperationActive()
    elif rv == CKR_OPERATION_NOT_INITIALIZED:
        exc = OperationNotInitialized()
    elif rv == CKR_PIN_EXPIRED:
        exc = PinExpired()
    elif rv == CKR_PIN_INCORRECT:
        exc = PinIncorrect()
    elif rv == CKR_PIN_INVALID:
        exc = PinInvalid()
    elif rv == CKR_PIN_LEN_RANGE:
        exc = PinLenRange()
    elif rv == CKR_PIN_LOCKED:
        exc = PinLocked()
    elif rv == CKR_PIN_TOO_WEAK:
        exc = PinTooWeak()
    elif rv == CKR_PUBLIC_KEY_INVALID:
        exc = PublicKeyInvalid()
    elif rv == CKR_RANDOM_NO_RNG:
        exc = RandomNoRNG()
    elif rv == CKR_RANDOM_SEED_NOT_SUPPORTED:
        exc = RandomSeedNotSupported()
    elif rv == CKR_SESSION_CLOSED:
        exc = SessionClosed()
    elif rv == CKR_SESSION_COUNT:
        exc = SessionCount()
    elif rv == CKR_SESSION_EXISTS:
        exc = SessionExists()
    elif rv == CKR_SESSION_HANDLE_INVALID:
        exc = SessionHandleInvalid()
    elif rv == CKR_SESSION_PARALLEL_NOT_SUPPORTED:
        exc = PKCS11Error("Parallel not supported. Should never see this.")
    elif rv == CKR_SESSION_READ_ONLY:
        exc = SessionReadOnly()
    elif rv == CKR_SESSION_READ_ONLY_EXISTS:
        exc = SessionReadOnlyExists()
    elif rv == CKR_SESSION_READ_WRITE_SO_EXISTS:
        exc = SessionReadWriteSOExists()
    elif rv == CKR_SIGNATURE_LEN_RANGE:
        exc = SignatureLenRange()
    elif rv == CKR_SIGNATURE_INVALID:
        exc = SignatureInvalid()
    elif rv == CKR_TEMPLATE_INCOMPLETE:
        exc = TemplateIncomplete()
    elif rv == CKR_TEMPLATE_INCONSISTENT:
        exc = TemplateInconsistent()
    elif rv == CKR_SLOT_ID_INVALID:
        exc = SlotIDInvalid()
    elif rv == CKR_TOKEN_NOT_PRESENT:
        exc = TokenNotPresent()
    elif rv == CKR_TOKEN_NOT_RECOGNIZED:
        exc = TokenNotRecognised()
    elif rv == CKR_TOKEN_WRITE_PROTECTED:
        exc = TokenWriteProtected()
    elif rv == CKR_UNWRAPPING_KEY_HANDLE_INVALID:
        exc = UnwrappingKeyHandleInvalid()
    elif rv == CKR_UNWRAPPING_KEY_SIZE_RANGE:
        exc = UnwrappingKeySizeRange()
    elif rv == CKR_UNWRAPPING_KEY_TYPE_INCONSISTENT:
        exc = UnwrappingKeyTypeInconsistent()
    elif rv == CKR_USER_NOT_LOGGED_IN:
        exc = UserNotLoggedIn()
    elif rv == CKR_USER_ALREADY_LOGGED_IN:
        exc = UserAlreadyLoggedIn()
    elif rv == CKR_USER_ANOTHER_ALREADY_LOGGED_IN:
        exc = AnotherUserAlreadyLoggedIn()
    elif rv == CKR_USER_PIN_NOT_INITIALIZED:
        exc = UserPinNotInitialized()
    elif rv == CKR_USER_TOO_MANY_TYPES:
        exc = UserTooManyTypes()
    elif rv == CKR_USER_TYPE_INVALID:
        exc = PKCS11Error("User type invalid. Should never see this.")
    elif rv == CKR_WRAPPED_KEY_INVALID:
        exc = WrappedKeyInvalid()
    elif rv == CKR_WRAPPED_KEY_LEN_RANGE:
        exc = WrappedKeyLenRange()
    elif rv == CKR_WRAPPING_KEY_HANDLE_INVALID:
        exc = WrappingKeyHandleInvalid()
    elif rv == CKR_WRAPPING_KEY_SIZE_RANGE:
        exc = WrappingKeySizeRange()
    elif rv == CKR_WRAPPING_KEY_TYPE_INCONSISTENT:
        exc = WrappingKeyTypeInconsistent()
    else:
        exc = PKCS11Error("Unmapped error code %s" % hex(rv))
    return exc



"""
Types for high level PKCS#11 wrapper.

This module provides stubs that are overrideen in pkcs11._pkcs11.
"""

from __future__ import annotations

from binascii import hexlify
from functools import cached_property
from typing import TYPE_CHECKING, Any, Final, Iterator

from pkcs11 import CancelStrategy
from pkcs11.constants import (
    Attribute,
    CertificateType,
    MechanismFlag,
    ObjectClass,
    SlotFlag,
    TokenFlag,
    UserType,
)
from pkcs11.exceptions import (
    ArgumentsBad,
    AttributeTypeInvalid,
    MultipleObjectsReturned,
    NoSuchKey,
    SignatureInvalid,
    SignatureLenRange,
)
from pkcs11.mechanisms import KeyType, Mechanism

if TYPE_CHECKING:
    from pkcs11.attributes import AttributeMapper

PROTECTED_AUTH: Final[object] = object()
"""Indicate the pin should be supplied via an external mechanism (e.g. pin pad)"""


class IdentifiedBy:
    __slots__ = ()

    def _identity(self) -> Any:
        raise NotImplementedError()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, IdentifiedBy) and self._identity() == other._identity()

    def __hash__(self) -> int:
        return hash(self._identity())


def _CK_UTF8CHAR_to_str(data: bytes) -> str:
    """
    Convert CK_UTF8CHAR to string.

    Substitutes invalid bytes with the replacement character to make usage more
    robust with non-compliant tokens.
    """
    return data.rstrip(b"\0").decode("utf-8", errors="replace").rstrip()


def _CK_VERSION_to_tuple(data: dict[str, int]) -> tuple[int, int]:
    """Convert CK_VERSION to tuple."""
    return (data["major"], data["minor"])


def _CK_MECHANISM_TYPE_to_enum(mechanism: int) -> Mechanism | int:
    """Convert CK_MECHANISM_TYPE to enum or be okay."""
    try:
        return Mechanism(mechanism)
    except ValueError:
        return mechanism


class MechanismInfo:
    """
    Information about a mechanism.

    See :meth:`pkcs11.Slot.get_mechanism_info`.
    """

    slot: Slot
    mechanism: Mechanism | int
    min_key_length: int | None
    max_key_length: int | None
    flags: MechanismFlag

    def __init__(
        self,
        slot: Slot,
        mechanism: Mechanism | int,
        ulMinKeySize: int | None = None,
        ulMaxKeySize: int | None = None,
        flags: int | None = None,
        **kwargs: Any,
    ) -> None:
        self.slot = slot
        """:class:`pkcs11.Slot` this information is for."""
        self.mechanism = mechanism
        """:class:`pkcs11.mechanisms.Mechanism` this information is for."""
        self.min_key_length = ulMinKeySize
        """Minimum key length in bits (:class:`int`)."""
        self.max_key_length = ulMaxKeySize
        """Maximum key length in bits (:class:`int`)."""
        self.flags = MechanismFlag(flags) if flags is not None else MechanismFlag(0)
        """Mechanism capabilities (:class:`pkcs11.constants.MechanismFlag`)."""

    def __str__(self) -> str:
        return "\n".join(
            (
                "Supported key lengths: [%s, %s]" % (self.min_key_length, self.max_key_length),
                "Flags: %s" % self.flags,
            )
        )

    def __repr__(self) -> str:
        return "<{klass} (mechanism={mechanism}, flags={flags})>".format(
            klass=type(self).__name__, mechanism=str(self.mechanism), flags=str(self.flags)
        )


class Slot(IdentifiedBy):
    """
    A PKCS#11 device slot.

    This object represents a physical or software slot exposed by PKCS#11.
    A slot has hardware capabilities, e.g. supported mechanisms and may has
    a physical or software :class:`Token` installed.
    """

    __slots__ = ()

    @property
    def flags(self) -> SlotFlag:
        """Capabilities of this slot (:class:`SlotFlag`)."""
        raise NotImplementedError()

    @property
    def hardware_version(self) -> tuple[int, int]:
        """Hardware version (:class:`tuple`)."""
        raise NotImplementedError()

    @property
    def firmware_version(self) -> tuple[int, int]:
        """Firmware version (:class:`tuple`)."""
        raise NotImplementedError()

    @property
    def cryptoki_version(self) -> tuple[int, int]:
        """PKCS#11 API version (:class: `tuple`)"""
        raise NotImplementedError()

    @property
    def slot_id(self) -> int:
        """Slot identifier (opaque)."""
        raise NotImplementedError()

    @property
    def slot_description(self) -> str:
        """Slot name (:class:`str`)."""
        raise NotImplementedError()

    @property
    def manufacturer_id(self) -> str:
        """Slot/device manufacturer's name (:class:`str`)."""
        raise NotImplementedError()

    def get_token(self) -> Token:
        """
        Returns the token loaded into this slot.

        :rtype: Token
        """
        raise NotImplementedError()

    def get_mechanisms(self) -> set[Mechanism]:
        """
        Returns the mechanisms supported by this device.

        :rtype: set(Mechanism)
        """
        raise NotImplementedError()

    def get_mechanism_info(self, mechanism: Mechanism) -> MechanismInfo:
        """
        Returns information about the mechanism.

        :param Mechanism mechanism: mechanism to learn about
        :rtype: MechanismInfo
        """
        raise NotImplementedError()


class Token(IdentifiedBy):
    """
    A PKCS#11 token.

    A token can be physically installed in a :class:`Slot`, or a software
    token, depending on your PKCS#11 library.
    """

    __slots__ = ()

    @property
    def flags(self) -> TokenFlag:
        """Capabilities of this token (:class:`TokenFlag`)."""
        raise NotImplementedError()

    @property
    def hardware_version(self) -> tuple[int, int]:
        """Hardware version (:class:`tuple`)."""
        raise NotImplementedError()

    @property
    def firmware_version(self) -> tuple[int, int]:
        """Firmware version (:class:`tuple`)."""
        raise NotImplementedError()

    @property
    def slot(self) -> Slot:
        """The :class:`Slot` this token is installed in."""
        raise NotImplementedError()

    @property
    def label(self) -> str:
        """Label of this token (:class:`str`)."""
        raise NotImplementedError()

    @property
    def serial(self) -> bytes:
        """Serial number of this token (:class:`bytes`)."""
        raise NotImplementedError()

    @property
    def manufacturer_id(self) -> str:
        """Manufacturer ID (:class:`str`)."""
        raise NotImplementedError()

    @property
    def model(self) -> str:
        """Model name (:class:`str`)."""
        raise NotImplementedError()

    def open(
        self,
        rw: bool = False,
        user_pin: str | bytes | object | None = None,
        so_pin: str | bytes | object | None = None,
        user_type: UserType | None = None,
        attribute_mapper: AttributeMapper | None = None,
        cancel_strategy: CancelStrategy = CancelStrategy.DEFAULT,
    ) -> Session:
        """
        Open a session on the token and optionally log in as a user or
        security officer (pass one of `user_pin` or `so_pin`). Pass PROTECTED_AUTH to
        indicate the pin should be supplied via an external mechanism (e.g. pin pad).

        Can be used as a context manager or close with :meth:`Session.close`.

        ::

            with token.open() as session:

                print(session)

        :param rw: True to create a read/write session.
        :param bytes user_pin: Authenticate to this session as a user.
        :param bytes so_pin: Authenticate to this session as a
            security officer.
        :param user_type: Sets the userType parameter to C_Login.
            Allows for vendor-defined values. Defaults to UserType.SO if
            so_pin is set, otherwise UserType.USER.
        :param attribute_mapper:
            Optionally pass in a custom :class:`pkcs11.attributes.AttributeMapper`.
        :param cancel_strategy:
            Cancellation strategy for interrupted cryptographic operations.

        :rtype: Session
        """
        raise NotImplementedError()


class Session(IdentifiedBy):
    """
    A PKCS#11 :class:`Token` session.

    A session is required to do nearly all operations on a token including
    encryption/signing/keygen etc.

    Create a session using :meth:`Token.open`. Sessions can be used as a
    context manager or closed with :meth:`close`.
    """

    __slots__ = ()

    def __enter__(self) -> Session:
        return self

    def __exit__(
        self,
        type_: type[BaseException] | None,
        value: BaseException | None,
        traceback: Any,
    ) -> None:
        self.close()

    @property
    def token(self) -> Token:
        """:class:`Token` this session is on."""
        raise NotImplementedError()

    @property
    def rw(self) -> bool:
        """True if this is a read/write session."""
        raise NotImplementedError()

    def close(self) -> None:
        """Close the session."""
        raise NotImplementedError()

    def reaffirm_credentials(self, pin: str | bytes) -> None:
        raise NotImplementedError()

    def get_key(
        self,
        object_class: ObjectClass | None = None,
        key_type: KeyType | None = None,
        label: str | None = None,
        id: bytes | None = None,
    ) -> Key:
        """
        Search for a key with any of `key_type`, `label` and/or `id`.

        Returns a single key or throws :class:`pkcs11.exceptions.NoSuchKey` or
        :class:`pkcs11.exceptions.MultipleObjectsReturned`.

        This is a simplified version of :meth:`get_objects`, which allows
        searching for any object.

        :param ObjectClass object_class: Optional object class.
        :param KeyType key_type: Optional key type.
        :param str label: Optional key label.
        :param bytes id: Optional key id.

        :rtype: Key
        """

        if object_class is None and key_type is None and label is None and id is None:
            raise ArgumentsBad("Must specify at least one search parameter.")

        attrs: dict[Attribute, Any] = {}

        if object_class is not None:
            attrs[Attribute.CLASS] = object_class

        if key_type is not None:
            attrs[Attribute.KEY_TYPE] = key_type

        if label is not None:
            attrs[Attribute.LABEL] = label

        if id is not None:
            attrs[Attribute.ID] = id

        iterator = self.get_objects(attrs)
        try:
            obj = next(iterator)
        except StopIteration as ex:
            raise NoSuchKey("No key matching %s" % attrs) from ex

        try:
            next(iterator)
            raise MultipleObjectsReturned("More than 1 key matches %s" % attrs)
        except StopIteration:
            pass
        # The caller expects a Key, but get_objects returns Object
        # This is a runtime assumption that the returned object is a Key
        return obj  # type: ignore[return-value]

    def get_objects(self, attrs: dict[Attribute, Any] | None = None) -> Iterator[Object]:
        """
        Search for objects matching `attrs`. Returns a generator.

        ::

            for obj in session.get_objects({
                Attribute.CLASS: ObjectClass.SECRET_KEY,
                Attribute.LABEL: 'MY LABEL',
            }):

                print(obj)

        This is the more generic version of :meth:`get_key`.

        :param dict(Attribute,*) attrs: Attributes to search for.

        :rtype: iter(Object)
        """
        raise NotImplementedError()

    def create_object(self, attrs: dict[Attribute, Any]) -> Object:
        """
        Create a new object on the :class:`Token`. This is a low-level
        interface to create any type of object and can be used for importing
        data onto the Token.

        ::

            key = session.create_object({
                pkcs11.Attribute.CLASS: pkcs11.ObjectClass.SECRET_KEY,
                pkcs11.Attribute.KEY_TYPE: pkcs11.KeyType.AES,
                pkcs11.Attribute.VALUE: b'SUPER SECRET KEY',
            })

        For generating keys see :meth:`generate_key` or
        :meth:`generate_keypair`.
        For importing keys see :ref:`importing-keys`.

        Requires a read/write session, unless the object is not to be
        stored. To permanently store the object in the HSM add **pkcs.Attribute.TOKEN: True**,
        see :meth:`pkcs11.Attribute` for more available object attributes.

        :param dict(Attribute,*) attrs: attributes of the object to create
        :rtype: Object
        """
        raise NotImplementedError()

    def create_domain_parameters(
        self,
        key_type: KeyType,
        attrs: dict[Attribute, Any],
        local: bool = False,
        store: bool = False,
    ) -> DomainParameters:
        """
        Create a domain parameters object from known parameters.

        Domain parameters are used for key generation of key types such
        as DH, DSA and EC.

        You can also generate new parameters using
        :meth:`generate_domain_parameters`.

        The `local` parameter creates a Python object that is not created on
        the HSM (its object handle will be unset). This is useful if you only
        need the domain parameters to create another object, and do not need a
        real PKCS #11 object in the session.

        .. warning::

            Domain parameters have no id or labels. Storing them is possible
            but be aware they may be difficult to retrieve.

        :param KeyType key_type: Key type these parameters are for
        :param dict(Attribute,*) attrs: Domain parameters
            (specific tp `key_type`)
        :param local: if True, do not transfer parameters to the HSM.
        :param store: if True, store these parameters permanently in the HSM.
        :rtype: DomainParameters
        """

        raise NotImplementedError()

    def generate_domain_parameters(
        self,
        key_type: KeyType,
        param_length: int,
        store: bool = False,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        template: dict[Attribute, Any] | None = None,
    ) -> DomainParameters:
        """
        Generate domain parameters.

        See :meth:`create_domain_parameters` for creating domain parameter
        objects from known parameters.

        See :meth:`generate_key` for documentation on mechanisms and templates.

        .. warning::

            Domain parameters have no id or labels. Storing them is possible
            but be aware they may be difficult to retrieve.

        :param KeyType key_type: Key type these parameters are for
        :param int param_length: Size of the parameters (e.g. prime length)
            in bits.
        :param store: Store these parameters in the HSM
        :param Mechanism mechanism: Optional generation mechanism (or default)
        :param bytes mechanism_param: Optional mechanism parameter.
        :param dict(Attribute,*) template: Optional additional attributes.
        :rtype: DomainParameters
        """

        raise NotImplementedError()

    def generate_key(
        self,
        key_type: KeyType,
        key_length: int | None = None,
        id: bytes | None = None,
        label: str | None = None,
        store: bool = False,
        capabilities: MechanismFlag | None = None,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        template: dict[Attribute, Any] | None = None,
    ) -> SecretKey:
        """
        Generate a single key (e.g. AES, DES).

        Keys should set at least `id` or `label`.

        An appropriate `mechanism` will be chosen for `key_type`
        (see :attr:`DEFAULT_GENERATE_MECHANISMS`) or this can be overridden.
        Similarly for the `capabilities` (see
        :attr:`DEFAULT_KEY_CAPABILITIES`).

        The `template` will extend the default template used to make the
        key.

        Possible mechanisms and template attributes are defined by `PKCS #11
        <http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/pkcs11-curr-v2.40.html>`_.
        Invalid mechanisms or attributes should raise
        :exc:`pkcs11.exceptions.MechanismInvalid` and
        :exc:`pkcs11.exceptions.AttributeTypeInvalid` respectively.

        :param KeyType key_type: Key type (e.g. KeyType.AES)
        :param int key_length: Key length in bits (e.g. 256).
        :param bytes id: Key identifier.
        :param str label: Key label.
        :param store: Store key on token (requires R/W session).
        :param MechanismFlag capabilities: Key capabilities (or default).
        :param Mechanism mechanism: Generation mechanism (or default).
        :param bytes mechanism_param: Optional vector to the mechanism.
        :param dict(Attribute,*) template: Additional attributes.

        :rtype: SecretKey
        """
        raise NotImplementedError()

    def generate_keypair(
        self,
        key_type: KeyType,
        key_length: int | None = None,
        **kwargs: Any,
    ) -> tuple[PublicKey, PrivateKey]:
        """
        Generate a asymmetric keypair (e.g. RSA).

        See :meth:`generate_key` for more information.

        :param KeyType key_type: Key type (e.g. KeyType.DSA)
        :param int key_length: Key length in bits (e.g. 256).
        :param bytes id: Key identifier.
        :param str label: Key label.
        :param bool store: Store key on token (requires R/W session).
        :param MechanismFlag capabilities: Key capabilities (or default).
        :param Mechanism mechanism: Generation mechanism (or default).
        :param bytes mechanism_param: Optional vector to the mechanism.
        :param dict(Attribute,*) private_template: Additional attributes for private key.
        :param dict(Attribute,*) public_template: Additional attributes for public key.

        :rtype: (PublicKey, PrivateKey)
        """
        if key_type is KeyType.DSA:
            if key_length is None:
                raise ArgumentsBad("Must specify `key_length`")

            params = self.generate_domain_parameters(key_type, key_length)

            return params.generate_keypair(**kwargs)
        else:
            return self._generate_keypair(key_type, key_length=key_length, **kwargs)

    def _generate_keypair(
        self,
        key_type: KeyType,
        key_length: int | None = None,
        **kwargs: Any,
    ) -> tuple[PublicKey, PrivateKey]:
        raise NotImplementedError()

    def seed_random(self, seed: bytes) -> None:
        """
        Mix additional seed material into the RNG (if supported).

        :param bytes seed: Bytes of random to seed.
        """
        raise NotImplementedError()

    def generate_random(self, nbits: int) -> bytes:
        """
        Generate `length` bits of random or pseudo-random data (if supported).

        :param int nbits: Number of bits to generate.
        :rtype: bytes
        """
        raise NotImplementedError()

    def digest(self, data: str | bytes | Key | Iterator[bytes | Key], **kwargs: Any) -> bytes:
        """
        Digest `data` using `mechanism`.

        `data` can be a single value or an iterator.

        :class:`Key` objects can also be digested, optionally interspersed
        with :class:`bytes`.

        :param data: Data to digest
        :type data: str, bytes, Key or iter(bytes, Key)
        :param Mechanism mechanism: digest mechanism
        :param bytes mechanism_param: optional mechanism parameter

        :rtype: bytes
        """

        # If data is a string, encode it now as UTF-8.
        if isinstance(data, str):
            data = data.encode("utf-8")

        if isinstance(data, bytes):
            return self._digest(data, **kwargs)

        elif isinstance(data, Key):
            return self._digest_generator(iter((data,)), **kwargs)

        return self._digest_generator(data, **kwargs)

    def set_pin(self, old_pin: str | bytes, new_pin: str | bytes) -> None:
        """Change the user pin."""
        raise NotImplementedError()

    def init_pin(self, pin: str | bytes) -> None:
        """
        Initializes the user PIN.

        Differs from set_pin in that it sets the user PIN for the first time.
        Once set, the pin can be changed using set_pin.
        """
        raise NotImplementedError()

    def _digest(
        self,
        data: bytes,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
    ) -> bytes:
        raise NotImplementedError()

    def _digest_generator(
        self,
        data: Iterator[bytes | Key],
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
    ) -> bytes:
        raise NotImplementedError()


class Object(IdentifiedBy):
    """
    A PKCS#11 object residing on a :class:`Token`.

    Objects implement :meth:`__getitem__` and :meth:`__setitem__` to
    retrieve :class:`pkcs11.constants.Attribute` values on the object.
    Valid attributes for an object are given in `PKCS #11
    <http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/pkcs11-curr-v2.40.html>`_.
    Invalid attributes should raise
    :exc:`pkcs11.exceptions.AttributeTypeInvalid`.
    """

    object_class: ObjectClass | None = None
    """:class:`pkcs11.constants.ObjectClass` of this Object."""

    @property
    def session(self) -> Session:
        raise NotImplementedError()

    @property
    def handle(self) -> int | None:
        raise NotImplementedError()

    def __getitem__(self, key: Attribute) -> Any:
        raise NotImplementedError()

    def __setitem__(self, key: Attribute, value: Any) -> None:
        raise NotImplementedError()

    def get_attributes(self, keys: list[Attribute]) -> dict[Attribute, Any]:
        raise NotImplementedError()

    def copy(self, attrs: dict[Attribute, Any]) -> Object:
        """
        Make a copy of the object with new attributes `attrs`.

        Requires a read/write session, unless the object is not to be
        stored.

        ::

            new = key.copy({
                Attribute.LABEL: 'MY NEW KEY',
            })

        Certain objects may not be copied. Calling :meth:`copy` on such
        objects will result in an exception.

        :param dict(Attribute,*) attrs: attributes for the new :class:`Object`
        :rtype: Object
        """
        raise NotImplementedError()

    def destroy(self) -> None:
        """
        Destroy the object.

        Requires a read/write session, unless the object is not stored.

        Certain objects may not be destroyed. Calling :meth:`destroy` on such
        objects will result in an exception.

        The :class:`Object` is no longer valid.
        """
        raise NotImplementedError()


class DomainParameters(Object):
    """
    PKCS#11 Domain Parameters.

    Used to store domain parameters as part of the key generation step, e.g.
    in DSA and Diffie-Hellman.
    """

    @cached_property
    def key_type(self) -> KeyType:
        """
        Key type (:class:`pkcs11.mechanisms.KeyType`) these parameters
        can be used to generate.
        """
        return self[Attribute.KEY_TYPE]

    def generate_keypair(
        self,
        id: bytes | None = None,
        label: str | None = None,
        store: bool = False,
        capabilities: MechanismFlag | None = None,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        public_template: dict[Attribute, Any] | None = None,
        private_template: dict[Attribute, Any] | None = None,
    ) -> tuple[PublicKey, PrivateKey]:
        """
        Generate a key pair from these domain parameters (e.g. for
        Diffie-Hellman.

        See :meth:`Session.generate_key` for more information.

        :param bytes id: Key identifier.
        :param str label: Key label.
        :param store: Store key on token (requires R/W session).
        :param MechanismFlag capabilities: Key capabilities (or default).
        :param Mechanism mechanism: Generation mechanism (or default).
        :param bytes mechanism_param: Optional vector to the mechanism.
        :param dict(Attribute,*) template: Additional attributes.

        :rtype: (PublicKey, PrivateKey)
        """
        raise NotImplementedError()


class LocalDomainParameters(DomainParameters):
    _session: Session
    params: dict[Attribute, Any]

    def __init__(self, session: Session, params: dict[Attribute, Any]) -> None:
        self._session = session
        self.params = params

    @property
    def session(self) -> Session:
        return self._session

    @property
    def handle(self) -> None:
        return None

    def __getitem__(self, key: Attribute) -> Any:
        try:
            return self.params[key]
        except KeyError as ex:
            raise AttributeTypeInvalid from ex

    def get_attributes(self, keys: list[Attribute]) -> dict[Attribute, Any]:
        return {key: self.params[key] for key in keys if key in self.params}

    def __setitem__(self, key: Attribute, value: Any) -> None:
        self.params[key] = value


class HasKeyType(Object):
    @cached_property
    def key_type(self) -> KeyType:
        """Key type (:class:`pkcs11.mechanisms.KeyType`)."""
        return self[Attribute.KEY_TYPE]


class Key(HasKeyType):
    """Base class for all key :class:`Object` types."""

    @property
    def key_length(self) -> int:
        """Key length in bits."""
        raise NotImplementedError

    @cached_property
    def id(self) -> bytes:
        """Key id (:class:`bytes`)."""
        return self[Attribute.ID]

    @cached_property
    def label(self) -> str:
        """Key label (:class:`str`)."""
        return self[Attribute.LABEL]

    @cached_property
    def _key_description(self) -> str:
        """A description of the key."""
        try:
            return "%s-bit %s" % (self.key_length, self.key_type.name)
        except AttributeTypeInvalid:
            return self.key_type.name

    def __repr__(self) -> str:
        return "<%s label='%s' id='%s' %s>" % (
            type(self).__name__,
            self.label,
            hexlify(self.id).decode("ascii"),
            self._key_description,
        )


class SecretKey(Key):
    """
    A PKCS#11 :attr:`pkcs11.constants.ObjectClass.SECRET_KEY` object
    (symmetric encryption key).
    """

    object_class: ObjectClass = ObjectClass.SECRET_KEY

    @cached_property
    def key_length(self) -> int:
        """Key length in bits."""
        return self[Attribute.VALUE_LEN] * 8


class PublicKey(Key):
    """
    A PKCS#11 :attr:`pkcs11.constants.ObjectClass.PUBLIC_KEY` object
    (asymmetric public key).

    RSA private keys can be imported and exported from PKCS#1 DER-encoding
    using :func:`pkcs11.util.rsa.decode_rsa_public_key` and
    :func:`pkcs11.util.rsa.encode_rsa_public_key` respectively.
    """

    object_class: ObjectClass = ObjectClass.PUBLIC_KEY

    @cached_property
    def key_length(self) -> int:
        """Key length in bits."""
        return self[Attribute.MODULUS_BITS]


class PrivateKey(Key):
    """
    A PKCS#11 :attr:`pkcs11.constants.ObjectClass.PRIVATE_KEY` object
    (asymmetric private key).

    RSA private keys can be imported from PKCS#1 DER-encoding using
    :func:`pkcs11.util.rsa.decode_rsa_private_key`.

    .. warning::

        Private keys imported directly, rather than unwrapped from a trusted
        private key should be considered insecure.
    """

    object_class: ObjectClass = ObjectClass.PRIVATE_KEY

    @cached_property
    def key_length(self) -> int:
        """Key length in bits."""
        return len(self[Attribute.MODULUS]) * 8


class Certificate(Object):
    """
    A PKCS#11 :attr:`pkcs11.constants.ObjectClass.CERTIFICATE` object.

    PKCS#11 is limited in its handling of certificates, and does not
    provide features like parsing of X.509 etc. These should be handled in
    an external library. PKCS#11 will not set attributes on the certificate
    based on the `VALUE`.

    :func:`pkcs11.util.x509.decode_x509_certificate` will extract attributes
    from a certificate to create the object.
    """

    object_class: ObjectClass = ObjectClass.CERTIFICATE

    @cached_property
    def certificate_type(self) -> CertificateType:
        """
        The type of certificate.

        :rtype: CertificateType
        """
        return self[Attribute.CERTIFICATE_TYPE]


class EncryptMixin(HasKeyType):
    """
    This :class:`Object` supports the encrypt capability.
    """

    def encrypt(
        self,
        data: str | bytes | Iterator[bytes],
        buffer_size: int = 8192,
        **kwargs: Any,
    ) -> bytes | Iterator[bytes]:
        """
        Encrypt some `data`.

        Data can be either :class:`str` or :class:`bytes`, in which case it
        will return :class:`bytes`; or an iterable of :class:`bytes` in
        which case it will return a generator yielding :class:`bytes`
        (be aware, more chunks will be output than input).

        If you do not specify `mechanism` then the default from
        :attr:`DEFAULT_ENCRYPT_MECHANISMS` will be used. If an iterable
        is passed and the mechanism chosen does not support handling data
        in chunks, an exception will be raised.

        Some mechanisms (including the default CBC mechanisms) require
        additional parameters, e.g. an initialisation vector [#]_, to
        the mechanism.  Pass this as `mechanism_param`.
        Documentation of these parameters is given specified in `PKCS #11
        <http://docs.oasis-open.org/pkcs11/pkcs11-curr/v2.40/pkcs11-curr-v2.40.html>`_.

        When passing an iterable for data
        `buffer_size` must be sufficient to store the working buffer. An
        integer number of blocks and greater than or equal to the largest
        input chunk is recommended.

        The returned generator obtains a lock on the :class:`Session`
        to prevent other threads from starting a simultaneous operation.
        The lock is released when you consume/destroy the generator.
        See :ref:`concurrency`.

        .. warning::

            It's not currently possible to cancel an encryption operation
            by deleting the generator. You must consume the generator to
            complete the operation.

        An example of streaming a file is as follows:

        ::

            def encrypt_file(file_in, file_out, buffer_size=8192):

                with \\
                        open(file_in, 'rb') as input_, \\
                        open(file_out, 'wb') as output:

                    chunks = iter(lambda: input_.read(buffer_size), '')

                    for chunk in key.encrypt(chunks,
                                             mechanism_param=iv,
                                             buffer_size=buffer_size):
                        output.write(chunk)

        :param data: data to encrypt
        :type data: str, bytes or iter(bytes)
        :param Mechanism mechanism: optional encryption mechanism
            (or None for default)
        :param bytes mechanism_param: optional mechanism parameter
            (e.g. initialisation vector).
        :param int buffer_size: size of the working buffer (for generators)

        :rtype: bytes or iter(bytes)

        .. [#] The initialisation vector should contain quality random,
            e.g. from :meth:`Session.generate_random`.
            This method will not return the value of the initialisation
            vector as part of the encryption. You must store that yourself.
        """

        # If data is a string, encode it now as UTF-8.
        if isinstance(data, str):
            data = data.encode("utf-8")

        if isinstance(data, bytes):
            return self._encrypt(data, **kwargs)

        else:
            return self._encrypt_generator(data, buffer_size=buffer_size, **kwargs)

    def _encrypt(
        self,
        data: bytes,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        buffer_size: int = 8192,
    ) -> bytes:
        raise NotImplementedError()

    def _encrypt_generator(
        self,
        data: Iterator[bytes],
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        buffer_size: int = 8192,
    ) -> Iterator[bytes]:
        raise NotImplementedError()


class DecryptMixin(HasKeyType):
    """
    This :class:`Object` supports the decrypt capability.
    """

    def decrypt(
        self,
        data: bytes | Iterator[bytes],
        buffer_size: int = 8192,
        **kwargs: Any,
    ) -> bytes | Iterator[bytes]:
        """
        Decrypt some `data`.

        See :meth:`EncryptMixin.encrypt` for more information.

        :param data: data to decrypt
        :type data: bytes or iter(bytes)
        :param Mechanism mechanism: optional encryption mechanism
            (or None for default).
        :param bytes mechanism_param: optional mechanism parameter
            (e.g. initialisation vector).
        :param pin: optional user pin for keys that require it (e.g. YubiKey)
        :param int buffer_size: size of the working buffer (for generators).

        :rtype: bytes or iter(bytes)
        """

        # If we're not an iterable, call into our generator with an iterable
        # version and join the result at the end.
        if isinstance(data, bytes):
            return self._decrypt(data, **kwargs)

        else:
            return self._decrypt_generator(data, buffer_size=buffer_size, **kwargs)

    def _decrypt(
        self,
        data: bytes,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        pin: str | bytes | None = None,
        buffer_size: int = 8192,
    ) -> bytes:
        raise NotImplementedError()

    def _decrypt_generator(
        self,
        data: Iterator[bytes],
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        pin: str | bytes | None = None,
        buffer_size: int = 8192,
    ) -> Iterator[bytes]:
        raise NotImplementedError()


class SignMixin(HasKeyType):
    """
    This :class:`Object` supports the sign capability.
    """

    def sign(self, data: str | bytes | Iterator[bytes], **kwargs: Any) -> bytes:
        """
        Sign some `data`.

        See :meth:`EncryptMixin.encrypt` for more information.

        For DSA and ECDSA keys, PKCS #11 outputs the two parameters (r & s)
        as two concatenated `biginteger` of the same length. To convert these
        into other formats, such as the format used by OpenSSL, use
        :func:`pkcs11.util.dsa.encode_dsa_signature` or
        :func:`pkcs11.util.ec.encode_ecdsa_signature`.

        :param data: data to sign
        :type data: str, bytes or iter(bytes)
        :param Mechanism mechanism: optional signing mechanism
        :param bytes mechanism_param: optional mechanism parameter
        :param pin: optional user pin for keys that require it (e.g. YubiKey)

        :rtype: bytes
        """

        # If data is a string, encode it now as UTF-8.
        if isinstance(data, str):
            data = data.encode("utf-8")

        if isinstance(data, bytes):
            return self._sign(data, **kwargs)

        else:
            return self._sign_generator(data, **kwargs)

    def _sign(
        self,
        data: bytes,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        pin: str | bytes | None = None,
        buffer_size: int = 8192,
    ) -> bytes:
        raise NotImplementedError()

    def _sign_generator(
        self,
        data: Iterator[bytes],
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        pin: str | bytes | None = None,
    ) -> bytes:
        raise NotImplementedError()


class VerifyMixin(HasKeyType):
    """
    This :class:`Object` supports the verify capability.
    """

    def verify(
        self,
        data: str | bytes | Iterator[bytes],
        signature: bytes,
        **kwargs: Any,
    ) -> bool:
        """
        Verify some `data`.

        See :meth:`EncryptMixin.encrypt` for more information.

        Returns True if `signature` is valid for `data`.

        For DSA and ECDSA keys, PKCS #11 expects the two parameters (r & s)
        as two concatenated `biginteger` of the same length. To convert these
        from other formats, such as the format used by OpenSSL, use
        :func:`pkcs11.util.dsa.decode_dsa_signature` or
        :func:`pkcs11.util.ec.decode_ecdsa_signature`.

        :param data: data to sign
        :type data: str, bytes or iter(bytes)
        :param bytes signature: signature
        :param Mechanism mechanism: optional signing mechanism
        :param bytes mechanism_param: optional mechanism parameter

        :rtype: bool
        """

        # If data is a string, encode it now as UTF-8.
        if isinstance(data, str):
            data = data.encode("utf-8")

        try:
            if isinstance(data, bytes):
                self._verify(data, signature, **kwargs)
            else:
                self._verify_generator(data, signature, **kwargs)

            return True

        except (SignatureInvalid, SignatureLenRange):
            return False

    def _verify(
        self,
        data: bytes,
        signature: bytes,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
    ) -> None:
        raise NotImplementedError()

    def _verify_generator(
        self,
        data: Iterator[bytes],
        signature: bytes,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
    ) -> None:
        raise NotImplementedError()


class WrapMixin(HasKeyType):
    """
    This :class:`Object` supports the wrap capability.
    """

    def wrap_key(
        self,
        key: Key,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
    ) -> bytes:
        """
        Use this key to wrap (i.e. encrypt) `key` for export. Returns
        an encrypted version of `key`.

        `key` must have :attr:`Attribute.EXTRACTABLE` = True.

        :param Key key: key to export
        :param Mechanism mechanism: wrapping mechanism (or None for default).
        :param bytes mechanism_param: mechanism parameter (if required)

        :rtype: bytes
        """
        raise NotImplementedError()


class UnwrapMixin(HasKeyType):
    """
    This :class:`Object` supports the unwrap capability.
    """

    def unwrap_key(
        self,
        object_class: ObjectClass,
        key_type: KeyType,
        key_data: bytes,
        id: bytes | None = None,
        label: str | None = None,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | None = None,
        store: bool = False,
        capabilities: MechanismFlag | None = None,
        template: dict[Attribute, Any] | None = None,
    ) -> Key:
        """
        Use this key to unwrap (i.e. decrypt) and import `key_data`.

        See :class:`Session.generate_key` for more information.

        :param ObjectClass object_class: Object class to import as
        :param KeyType key_type: Key type (e.g. KeyType.AES)
        :param bytes key_data: Encrypted key to unwrap
        :param bytes id: Key identifier.
        :param str label: Key label.
        :param store: Store key on token (requires R/W session).
        :param MechanismFlag capabilities: Key capabilities (or default).
        :param Mechanism mechanism: Generation mechanism (or default).
        :param bytes mechanism_param: Optional vector to the mechanism.
        :param dict(Attribute,*) template: Additional attributes.

        :rtype: Key
        """
        raise NotImplementedError()


class DeriveMixin(HasKeyType):
    """
    This :class:`Object` supports the derive capability.
    """

    def derive_key(
        self,
        key_type: KeyType,
        key_length: int,
        id: bytes | None = None,
        label: str | None = None,
        store: bool = False,
        capabilities: MechanismFlag | None = None,
        mechanism: Mechanism | None = None,
        mechanism_param: bytes | tuple[Any, ...] | None = None,
        template: dict[Attribute, Any] | None = None,
    ) -> SecretKey:
        """
        Derive a new key from this key. Used to create session
        keys from a PKCS key exchange.

        Typically the mechanism, e.g. Diffie-Hellman, requires you
        to specify the other party's piece of shared information as
        the `mechanism_param`.  Some mechanisms require a tuple of data (see
        :class:`pkcs11.mechanisms.Mechanism`).

        See :class:`Session.generate_key` for more documentation on key
        generation.

        Diffie-Hellman example:

        ::

            # Diffie-Hellman domain parameters
            # e.g. from RFC 3526, RFC 5114 or `openssl dhparam`
            prime = [0xFF, ...]
            base = [0x02]

            parameters = session.create_domain_parameters(KeyType.DH, {
                Attribute.PRIME: prime,
                Attribute.BASE: base,
            }, local=True)

            # Alice generates a DH key pair from the public
            # Diffie-Hellman parameters
            public, private = parameters.generate_keypair()
            alices_value = public[Attribute.VALUE]

            # Bob generates a DH key pair from the same parameters.

            # Alice exchanges public values with Bob...
            # She sends `alices_value` and receives `bobs_value`.
            # (Assuming Alice is doing AES CBC, she also needs to send an IV)

            # Alice generates a session key with Bob's public value
            # Bob will generate the same session key using Alice's value.
            session_key = private.derive_key(
                KeyType.AES, 128,
                mechanism_param=bobs_value)

        Elliptic-Curve Diffie-Hellman example:

        ::

            # DER encoded EC params, e.g. from OpenSSL
            # openssl ecparam -outform der -name prime192v1 | base64
            #
            # Check what EC parameters the module supports with
            # slot.get_module_info()
            parameters = session.create_domain_parameters(KeyType.EC, {
                Attribute.EC_PARAMS: b'...',
            }, local=True)

            # Alice generates a EC key pair, and gets her public value
            public, private = parameters.generate_keypair()
            alices_value = public[Attribute.EC_POINT]

            # Bob generates a DH key pair from the same parameters.

            # Alice exchanges public values with Bob...
            # She sends `alices_value` and receives `bobs_value`.

            # Alice generates a session key with Bob's public value
            # Bob will generate the same session key using Alice's value.
            session_key = private.derive_key(
                KeyType.AES, 128,
                mechanism_param=(KDF.NULL, None, bobs_value))

        :param KeyType key_type: Key type (e.g. KeyType.AES)
        :param int key_length: Key length in bits (e.g. 256).
        :param bytes id: Key identifier.
        :param str label: Key label.
        :param store: Store key on token (requires R/W session).
        :param MechanismFlag capabilities: Key capabilities (or default).
        :param Mechanism mechanism: Generation mechanism (or default).
        :param bytes mechanism_param: Optional vector to the mechanism.
        :param dict(Attribute,*) template: Additional attributes.

        :rtype: SecretKey
        """
        raise NotImplementedError()



from __future__ import annotations

from datetime import datetime
from enum import IntEnum
from struct import Struct
from typing import Any, Callable, Final

from pkcs11.constants import Attribute, CertificateType, MechanismFlag, ObjectClass
from pkcs11.mechanisms import KeyType, Mechanism

# Type aliases for pack/unpack function pairs
PackFunc = Callable[[Any], bytes]
UnpackFunc = Callable[[bytes], Any]
Handler = tuple[PackFunc, UnpackFunc]

# (Pack Function, Unpack Function) functions
_bool_struct = Struct("?")
_ulong_struct = Struct("L")

handle_bool: Handler = (
    _bool_struct.pack,
    lambda v: False if len(v) == 0 else _bool_struct.unpack(v)[0],
)
handle_ulong: Handler = (_ulong_struct.pack, lambda v: _ulong_struct.unpack(v)[0])
handle_str: Handler = (lambda s: s.encode("utf-8"), lambda b: b.decode("utf-8"))
handle_date: Handler = (
    lambda s: s.strftime("%Y%m%d").encode("ascii"),
    lambda s: datetime.strptime(s.decode("ascii"), "%Y%m%d").date(),
)
handle_bytes: Handler = (bytes, bytes)
# The PKCS#11 biginteger type is an array of bytes in network byte order.
# If you have an int type, wrap it in biginteger()
handle_biginteger: Handler = handle_bytes


def _enum(type_: type[IntEnum]) -> Handler:
    """Factory to pack/unpack ints into IntEnums."""
    pack, unpack = handle_ulong

    return (lambda v: pack(int(v)), lambda v: type_(unpack(v)))


ATTRIBUTE_TYPES: dict[Attribute, Handler] = {
    Attribute.ALWAYS_AUTHENTICATE: handle_bool,
    Attribute.ALWAYS_SENSITIVE: handle_bool,
    Attribute.APPLICATION: handle_str,
    Attribute.BASE: handle_biginteger,
    Attribute.CERTIFICATE_TYPE: _enum(CertificateType),
    Attribute.CHECK_VALUE: handle_bytes,
    Attribute.CLASS: _enum(ObjectClass),
    Attribute.COEFFICIENT: handle_biginteger,
    Attribute.DECRYPT: handle_bool,
    Attribute.DERIVE: handle_bool,
    Attribute.EC_PARAMS: handle_bytes,
    Attribute.EC_POINT: handle_bytes,
    Attribute.ENCRYPT: handle_bool,
    Attribute.END_DATE: handle_date,
    Attribute.EXPONENT_1: handle_biginteger,
    Attribute.EXPONENT_2: handle_biginteger,
    Attribute.EXTRACTABLE: handle_bool,
    Attribute.HASH_OF_ISSUER_PUBLIC_KEY: handle_bytes,
    Attribute.HASH_OF_SUBJECT_PUBLIC_KEY: handle_bytes,
    Attribute.ID: handle_bytes,
    Attribute.ISSUER: handle_bytes,
    Attribute.KEY_GEN_MECHANISM: _enum(Mechanism),
    Attribute.KEY_TYPE: _enum(KeyType),
    Attribute.LABEL: handle_str,
    Attribute.LOCAL: handle_bool,
    Attribute.MODIFIABLE: handle_bool,
    Attribute.COPYABLE: handle_bool,
    Attribute.MODULUS: handle_biginteger,
    Attribute.MODULUS_BITS: handle_ulong,
    Attribute.NEVER_EXTRACTABLE: handle_bool,
    Attribute.OBJECT_ID: handle_bytes,
    Attribute.PRIME: handle_biginteger,
    Attribute.PRIME_BITS: handle_ulong,
    Attribute.PRIME_1: handle_biginteger,
    Attribute.PRIME_2: handle_biginteger,
    Attribute.PRIVATE: handle_bool,
    Attribute.PRIVATE_EXPONENT: handle_biginteger,
    Attribute.PUBLIC_EXPONENT: handle_biginteger,
    Attribute.SENSITIVE: handle_bool,
    Attribute.SERIAL_NUMBER: handle_bytes,
    Attribute.SIGN: handle_bool,
    Attribute.SIGN_RECOVER: handle_bool,
    Attribute.START_DATE: handle_date,
    Attribute.SUBJECT: handle_bytes,
    Attribute.SUBPRIME: handle_biginteger,
    Attribute.SUBPRIME_BITS: handle_ulong,
    Attribute.TOKEN: handle_bool,
    Attribute.TRUSTED: handle_bool,
    Attribute.UNIQUE_ID: handle_str,
    Attribute.UNWRAP: handle_bool,
    Attribute.URL: handle_str,
    Attribute.VALUE: handle_biginteger,
    Attribute.VALUE_BITS: handle_ulong,
    Attribute.VALUE_LEN: handle_ulong,
    Attribute.VERIFY: handle_bool,
    Attribute.VERIFY_RECOVER: handle_bool,
    Attribute.WRAP: handle_bool,
    Attribute.WRAP_WITH_TRUSTED: handle_bool,
    Attribute.GOSTR3410_PARAMS: handle_bytes,
    Attribute.GOSTR3411_PARAMS: handle_bytes,
}
"""
Map of attributes to (serialize, deserialize) functions.
"""

ALL_CAPABILITIES: Final[tuple[Attribute, ...]] = (
    Attribute.ENCRYPT,
    Attribute.DECRYPT,
    Attribute.WRAP,
    Attribute.UNWRAP,
    Attribute.SIGN,
    Attribute.VERIFY,
    Attribute.DERIVE,
)


def _apply_common(
    template: dict[Attribute, Any],
    id_: bytes | None,
    label: str | None,
    store: bool,
) -> None:
    if id_:
        template[Attribute.ID] = id_
    if label:
        template[Attribute.LABEL] = label
    template[Attribute.TOKEN] = bool(store)


def _apply_capabilities(
    template: dict[Attribute, Any],
    possible_capas: tuple[Attribute, ...],
    capabilities: MechanismFlag | int,
) -> None:
    for attr in possible_capas:
        template[attr] = _capa_attr_to_mechanism_flag[attr] & capabilities


_capa_attr_to_mechanism_flag: Final[dict[Attribute, MechanismFlag]] = {
    Attribute.ENCRYPT: MechanismFlag.ENCRYPT,
    Attribute.DECRYPT: MechanismFlag.DECRYPT,
    Attribute.WRAP: MechanismFlag.WRAP,
    Attribute.UNWRAP: MechanismFlag.UNWRAP,
    Attribute.SIGN: MechanismFlag.SIGN,
    Attribute.VERIFY: MechanismFlag.VERIFY,
    Attribute.DERIVE: MechanismFlag.DERIVE,
}


class AttributeMapper:
    """
    Class mapping PKCS#11 attributes to and from Python values.
    """

    attribute_types: dict[Attribute, Handler]
    default_secret_key_template: dict[Attribute, Any]
    default_public_key_template: dict[Attribute, Any]
    default_private_key_template: dict[Attribute, Any]

    def __init__(self) -> None:
        self.attribute_types = dict(ATTRIBUTE_TYPES)
        self.default_secret_key_template = {
            Attribute.CLASS: ObjectClass.SECRET_KEY,
            Attribute.ID: b"",
            Attribute.LABEL: "",
            Attribute.PRIVATE: True,
            Attribute.SENSITIVE: True,
        }
        self.default_public_key_template = {
            Attribute.CLASS: ObjectClass.PUBLIC_KEY,
            Attribute.ID: b"",
            Attribute.LABEL: "",
        }
        self.default_private_key_template = {
            Attribute.CLASS: ObjectClass.PRIVATE_KEY,
            Attribute.ID: b"",
            Attribute.LABEL: "",
            Attribute.PRIVATE: True,
            Attribute.SENSITIVE: True,
        }

    def register_handler(self, key: Attribute, pack: PackFunc, unpack: UnpackFunc) -> None:
        self.attribute_types[key] = (pack, unpack)

    def _handler(self, key: Attribute) -> Handler:
        try:
            return self.attribute_types[key]
        except KeyError as e:
            raise NotImplementedError(f"Can't handle attribute type {hex(key)}.") from e

    def pack_attribute(self, key: Attribute, value: Any) -> bytes:
        """Pack a Attribute value into a bytes array."""
        pack, _ = self._handler(key)
        return pack(value)

    def unpack_attributes(self, key: Attribute, value: bytes) -> Any:
        """Unpack a Attribute bytes array into a Python value."""
        _, unpack = self._handler(key)
        return unpack(value)

    def public_key_template(
        self,
        *,
        capabilities: MechanismFlag | int,
        id_: bytes | None,
        label: str | None,
        store: bool,
    ) -> dict[Attribute, Any]:
        template = dict(self.default_public_key_template)
        _apply_capabilities(
            template, (Attribute.ENCRYPT, Attribute.WRAP, Attribute.VERIFY), capabilities
        )
        _apply_common(template, id_, label, store)
        return template

    def private_key_template(
        self,
        *,
        capabilities: MechanismFlag | int,
        id_: bytes | None,
        label: str | None,
        store: bool,
    ) -> dict[Attribute, Any]:
        template = dict(self.default_private_key_template)
        _apply_capabilities(
            template,
            (Attribute.DECRYPT, Attribute.UNWRAP, Attribute.SIGN, Attribute.DERIVE),
            capabilities,
        )
        _apply_common(template, id_, label, store)
        return template

    def secret_key_template(
        self,
        *,
        capabilities: MechanismFlag | int,
        id_: bytes | None,
        label: str | None,
        store: bool,
    ) -> dict[Attribute, Any]:
        return self.generic_key_template(
            self.default_secret_key_template,
            capabilities=capabilities,
            id_=id_,
            label=label,
            store=store,
        )

    def generic_key_template(
        self,
        base_template: dict[Attribute, Any],
        *,
        capabilities: MechanismFlag | int,
        id_: bytes | None,
        label: str | None,
        store: bool,
    ) -> dict[Attribute, Any]:
        template = dict(base_template)
        _apply_capabilities(template, ALL_CAPABILITIES, capabilities)
        _apply_common(template, id_, label, store)
        return template



"""
Key handling utilities for Diffie-Hellman keys.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from asn1crypto.algos import DHParameters
from asn1crypto.core import Integer

from pkcs11.constants import Attribute
from pkcs11.util import biginteger

if TYPE_CHECKING:
    from pkcs11.types import DomainParameters, PublicKey


def decode_dh_domain_parameters(der: bytes) -> dict[Attribute, Any]:
    """
    Decode DER-encoded Diffie-Hellman domain parameters.

    :param bytes der: DER-encoded parameters
    :rtype: dict(Attribute,*)
    """

    params = DHParameters.load(der)

    return {
        Attribute.BASE: biginteger(params["g"]),
        Attribute.PRIME: biginteger(params["p"]),
    }


def encode_dh_domain_parameters(obj: DomainParameters) -> bytes:
    """
    Encode DH domain parameters into DER-encoded format.

    Calculates the subprime if it isn't available.

    :param DomainParameters obj: domain parameters
    :rtype: bytes
    """

    asn1 = DHParameters(
        {
            "g": int.from_bytes(obj[Attribute.BASE], byteorder="big"),
            "p": int.from_bytes(obj[Attribute.PRIME], byteorder="big"),
        }
    )

    return asn1.dump()


def encode_dh_public_key(key: PublicKey) -> bytes:
    """
    Encode DH public key into RFC 3279 DER-encoded format.

    :param PublicKey key: public key
    :rtype: bytes
    """

    asn1 = Integer(int.from_bytes(key[Attribute.VALUE], byteorder="big"))

    return asn1.dump()


def decode_dh_public_key(der: bytes) -> bytes:
    """
    Decode a DH public key from RFC 3279 DER-encoded format.

    Returns a `biginteger` encoded as bytes.

    :param bytes der: DER-encoded public key
    :rtype: bytes
    """

    asn1 = Integer.load(der)
    return biginteger(asn1)
